@@ -23,8 +23,14 @@ Blind reimplementation of WebSockets as a standalone wrapper for Twisted
 protocols.
 """
 
+from __future__ import division
+
 __version__ = "0.7.1"
 
+import six
+
+import array
+
 from base64 import b64encode, b64decode
 from hashlib import md5, sha1
 from string import digits
@@ -130,10 +136,10 @@ def complete_hybi00(headers, challenge):
     key1 = headers["Sec-WebSocket-Key1"]
     key2 = headers["Sec-WebSocket-Key2"]
 
-    first = int("".join(i for i in key1 if i in digits)) / key1.count(" ")
-    second = int("".join(i for i in key2 if i in digits)) / key2.count(" ")
+    first = int("".join(i for i in key1 if i in digits)) // key1.count(" ")
+    second = int("".join(i for i in key2 if i in digits)) // key2.count(" ")
 
-    nonce = pack(">II8s", first, second, challenge)
+    nonce = pack(">II8s", first, second, six.b(challenge))
 
     return md5(nonce).digest()
 
@@ -146,7 +152,10 @@ def make_accept(key):
 
     guid = "258EAFA5-E914-47DA-95CA-C5AB0DC85B11"
 
-    return sha1("%s%s" % (key, guid)).digest().encode("base64").strip()
+    accept = "%s%s" % (key, guid)
+    hashed_bytes = sha1(accept.encode('utf-8')).digest()
+
+    return b64encode(hashed_bytes).strip().decode('utf-8')
 
 # Frame helpers.
 # Separated out to make unit testing a lot easier.
@@ -160,7 +169,10 @@ def make_hybi00_frame(buf):
     and valid text without any 0xff bytes.
     """
 
-    return "\x00%s\xff" % buf
+    if isinstance(buf, six.text_type):
+        buf = buf.encode('utf-8')
+
+    return six.b("\x00") + buf + six.b("\xff")
 
 def parse_hybi00_frames(buf):
     """
@@ -170,12 +182,12 @@ def parse_hybi00_frames(buf):
     and will actively ignore it.
     """
 
-    start = buf.find("\x00")
+    start = buf.find(six.b("\x00"))
     tail = 0
     frames = []
 
     while start != -1:
-        end = buf.find("\xff", start + 1)
+        end = buf.find(six.b("\xff"), start + 1)
         if end == -1:
             # Incomplete frame, try again later.
             break
@@ -184,7 +196,7 @@ def parse_hybi00_frames(buf):
             frame = buf[start + 1:end]
             frames.append((NORMAL, frame))
             tail = end + 1
-        start = buf.find("\x00", end + 1)
+        start = buf.find(six.b("\x00"), end + 1)
 
     # Adjust the buffer and return.
     buf = buf[tail:]
@@ -197,12 +209,12 @@ def mask(buf, key):
     The key must be exactly four bytes long.
     """
 
-    # This is super-secure, I promise~
-    key = [ord(i) for i in key]
-    buf = list(buf)
-    for i, char in enumerate(buf):
-        buf[i] = chr(ord(char) ^ key[i % 4])
-    return "".join(buf)
+    # This is super-duper-secure, I promise~
+    key = array.array("B", key)
+    buf = array.array("B", buf)
+    for i in range(len(buf)):
+        buf[i] ^= key[i % 4]
+    return buf.tostring()
 
 def make_hybi07_frame(buf, opcode=0x1):
     """
@@ -219,10 +231,25 @@ def make_hybi07_frame(buf, opcode=0x1):
     else:
         length = chr(len(buf))
 
+    if isinstance(buf, six.text_type):
+        buf = buf.encode('utf-8')
+
     # Always make a normal packet.
     header = chr(0x80 | opcode)
-    frame = "%s%s%s" % (header, length, buf)
-    return frame
+    return six.b(header + length) + buf
+
+def make_hybi07_frame_dwim(buf):
+    """
+    Make a HyBi-07 frame with binary or text data according to the type of buf.
+    """
+
+    # TODO: eliminate magic numbers.
+    if isinstance(buf, six.binary_type):
+        return make_hybi07_frame(buf, opcode=0x2)
+    elif isinstance(buf, six.text_type):
+        return make_hybi07_frame(buf.encode("utf-8"), opcode=0x1)
+    else:
+        raise TypeError("In binary support mode, frame data must be either str or unicode")
 
 def parse_hybi07_frames(buf):
     """
@@ -239,7 +266,11 @@ def parse_hybi07_frames(buf):
 
         # Grab the header. This single byte holds some flags nobody cares
         # about, and an opcode which nobody cares about.
-        header = ord(buf[start])
+        header = buf[start]
+
+        if six.PY2:
+            header = ord(header)
+
         if header & 0x70:
             # At least one of the reserved flags is set. Pork chop sandwiches!
             raise WSException("Reserved flag in HyBi-07 frame (%d)" % header)
@@ -256,7 +287,11 @@ def parse_hybi07_frames(buf):
 
         # Get the payload length and determine whether we need to look for an
         # extra length.
-        length = ord(buf[start + 1])
+        length = buf[start + 1]
+
+        if six.PY2:
+            length = ord(length)
+
         masked = length & 0x80
         length &= 0x7f
 
@@ -307,7 +342,7 @@ def parse_hybi07_frames(buf):
                 data = unpack(">H", data[:2])[0], data[2:]
             else:
                 # No reason given; use generic data.
-                data = 1000, "No reason given"
+                data = 1000, six.b("No reason given")
 
         frames.append((opcode, data))
         start += offset + length
@@ -320,18 +355,27 @@ class WebSocketProtocol(ProtocolWrapper):
     layer.
     """
 
-    buf = ""
+    buf = six.b("")
     codec = None
     location = "/"
     host = "example.com"
     origin = "http://example.com"
     state = REQUEST
     flavor = None
+    do_binary_frames = False
 
     def __init__(self, *args, **kwargs):
         ProtocolWrapper.__init__(self, *args, **kwargs)
         self.pending_frames = []
 
+    def setBinaryMode(self, mode):
+        """
+        If True, send str as binary and unicode as text.
+
+        Defaults to false for backwards compatibility.
+        """
+        self.do_binary_frames = bool(mode)
+
     def isSecure(self):
         """
         Borrowed technique for determining whether this connection is over
@@ -340,6 +384,14 @@ class WebSocketProtocol(ProtocolWrapper):
 
         return ISSLTransport(self.transport, None) is not None
 
+    def writeEncoded(self, data):
+        if isinstance(data, six.text_type):
+            data = data.encode('utf-8')
+        self.transport.write(data)
+
+    def writeEncodedSequence(self, sequence):
+        self.transport.writeSequence([ele.encode('utf-8') for ele in sequence])
+
     def sendCommonPreamble(self):
         """
         Send the preamble common to all WebSockets connections.
@@ -347,7 +399,7 @@ class WebSocketProtocol(ProtocolWrapper):
         This might go away in the future if WebSockets continue to diverge.
         """
 
-        self.transport.writeSequence([
+        self.writeEncodedSequence([
             "HTTP/1.1 101 FYI I am not a webserver\r\n",
             "Server: TwistedWebSocketWrapper/1.0\r\n",
             "Date: %s\r\n" % datetimeToString(),
@@ -364,7 +416,7 @@ class WebSocketProtocol(ProtocolWrapper):
 
         self.sendCommonPreamble()
 
-        self.transport.writeSequence([
+        self.writeEncodedSequence([
             "Sec-WebSocket-Origin: %s\r\n" % self.origin,
             "Sec-WebSocket-Location: %s://%s%s\r\n" % (protocol, self.host,
                                                        self.location),
@@ -379,10 +431,14 @@ class WebSocketProtocol(ProtocolWrapper):
         """
 
         self.sendCommonPreamble()
+
+        if self.codec:
+            self.transport.write("Sec-WebSocket-Protocol: %s\r\n" % self.codec)
+
         challenge = self.headers["Sec-WebSocket-Key"]
         response = make_accept(challenge)
 
-        self.transport.write("Sec-WebSocket-Accept: %s\r\n\r\n" % response)
+        self.writeEncoded("Sec-WebSocket-Accept: %s\r\n\r\n" % response)
 
     def parseFrames(self):
         """
@@ -398,7 +454,7 @@ class WebSocketProtocol(ProtocolWrapper):
 
         try:
             frames, self.buf = parser(self.buf)
-        except WSException, wse:
+        except WSException as wse:
             # Couldn't parse all the frames, something went wrong, let's bail.
             self.close(wse.args[0])
             return
@@ -430,7 +486,10 @@ class WebSocketProtocol(ProtocolWrapper):
         if self.flavor == HYBI00:
             maker = make_hybi00_frame
         elif self.flavor in (HYBI07, HYBI10, RFC6455):
-            maker = make_hybi07_frame
+            if self.do_binary_frames:
+                maker = make_hybi07_frame_dwim
+            else:
+                maker = make_hybi07_frame
         else:
             raise WSException("Unknown flavor %r" % self.flavor)
 
@@ -439,7 +498,7 @@ class WebSocketProtocol(ProtocolWrapper):
             if self.codec:
                 frame = encoders[self.codec](frame)
             packet = maker(frame)
-            self.transport.write(packet)
+            self.writeEncoded(packet)
         self.pending_frames = []
 
     def validateHeaders(self):
@@ -460,18 +519,28 @@ class WebSocketProtocol(ProtocolWrapper):
             self.origin = self.headers["Origin"]
 
         # Check whether a codec is needed. WS calls this a "protocol" for
-        # reasons I cannot fathom.
-        protocol = None
+        # reasons I cannot fathom. Newer versions of noVNC (0.4+) sets
+        # multiple comma-separated codecs, handle this by chosing first one
+        # we can encode/decode.
+        protocols = None
         if "WebSocket-Protocol" in self.headers:
-            protocol = self.headers["WebSocket-Protocol"]
+            protocols = self.headers["WebSocket-Protocol"]
         elif "Sec-WebSocket-Protocol" in self.headers:
-            protocol = self.headers["Sec-WebSocket-Protocol"]
+            protocols = self.headers["Sec-WebSocket-Protocol"]
+
+        if isinstance(protocols, six.string_types):
+            protocols = [p.strip() for p in protocols.split(',')]
+
+            for protocol in protocols:
+                if protocol in encoders or protocol in decoders:
+                    log.msg("Using WS protocol %s!" % protocol)
+                    self.codec = protocol
+                    break
 
-        if protocol:
-            if protocol not in encoders or protocol not in decoders:
                 log.msg("Couldn't handle WS protocol %s!" % protocol)
+
+            if not self.codec:
                 return False
-            self.codec = protocol
 
         # Start the next phase of the handshake for HyBi-00.
         if is_hybi00(self.headers):
@@ -518,8 +587,11 @@ class WebSocketProtocol(ProtocolWrapper):
             # These lines look like:
             # GET /some/path/to/a/websocket/resource HTTP/1.1
             if self.state == REQUEST:
-                if "\r\n" in self.buf:
-                    request, chaff, self.buf = self.buf.partition("\r\n")
+                separator = six.b("\r\n")
+                if separator in self.buf:
+                    request, chaff, self.buf = self.buf.partition(separator)
+                    request = request.decode('utf-8')
+
                     try:
                         verb, self.location, version = request.split(" ")
                     except ValueError:
@@ -529,8 +601,11 @@ class WebSocketProtocol(ProtocolWrapper):
 
             elif self.state == NEGOTIATING:
                 # Check to see if we've got a complete set of headers yet.
-                if "\r\n\r\n" in self.buf:
-                    head, chaff, self.buf = self.buf.partition("\r\n\r\n")
+                separator = six.b("\r\n\r\n")
+                if separator in self.buf:
+                    head, chaff, self.buf = self.buf.partition(separator)
+                    head = head.decode('utf-8')
+
                     self.headers = http_headers(head)
                     # Validate headers. This will cause a state change.
                     if not self.validateHeaders():
@@ -541,9 +616,11 @@ class WebSocketProtocol(ProtocolWrapper):
                 # HyBi-00/Hixie-76.
                 if len(self.buf) >= 8:
                     challenge, self.buf = self.buf[:8], self.buf[8:]
+                    challenge = challenge.decode('utf-8')
+
                     response = complete_hybi00(self.headers, challenge)
                     self.sendHyBi00Preamble()
-                    self.transport.write(response)
+                    self.writeEncoded(response)
                     log.msg("Completed HyBi-00/Hixie-76 handshake")
                     # We're all finished here; start sending frames.
                     self.state = FRAMES
@@ -595,7 +672,7 @@ class WebSocketProtocol(ProtocolWrapper):
         # from hanging.)
         if self.flavor in (HYBI07, HYBI10, RFC6455):
             frame = make_hybi07_frame(reason, opcode=0x8)
-            self.transport.write(frame)
+            self.writeEncoded(frame)
 
         self.loseConnection()
 
